@@ -146,7 +146,7 @@ conn_source(connObject *self, PyObject *noargs)
     }
 
     /* allocates new query object */
-    if (!(source_obj = PyObject_NEW(sourceObject, &sourceType))) {
+    if (!(source_obj = PyObject_New(sourceObject, &sourceType))) {
         return NULL;
     }
 
@@ -160,9 +160,57 @@ conn_source(connObject *self, PyObject *noargs)
     return (PyObject *) source_obj;
 }
 
-/* Base method for execution of both unprepared and prepared queries */
+/* For a non-query result, set the appropriate error status,
+   return the appropriate value, and free the result set. */
 static PyObject *
-_conn_query(connObject *self, PyObject *args, int prepared)
+_conn_non_query_result(int status, PGresult* result, PGconn *cnx)
+{
+    switch (status) {
+        case PGRES_EMPTY_QUERY:
+            PyErr_SetString(PyExc_ValueError, "Empty query");
+            break;
+        case PGRES_BAD_RESPONSE:
+        case PGRES_FATAL_ERROR:
+        case PGRES_NONFATAL_ERROR:
+            set_error(ProgrammingError, "Cannot execute query",
+                cnx, result);
+            break;
+        case PGRES_COMMAND_OK:
+            {   /* INSERT, UPDATE, DELETE */
+                Oid oid = PQoidValue(result);
+
+                if (oid == InvalidOid) {  /* not a single insert */
+                    char *ret = PQcmdTuples(result);
+
+                    if (ret[0]) {  /* return number of rows affected */
+                        PyObject *obj = PyStr_FromString(ret);
+                        PQclear(result);
+                        return obj;
+                    }
+                    PQclear(result);
+                    Py_INCREF(Py_None);
+                    return Py_None;
+                }
+                /* for a single insert, return the oid */
+                PQclear(result);
+                return PyInt_FromLong((long) oid);
+            }
+        case PGRES_COPY_OUT: /* no data will be received */
+        case PGRES_COPY_IN:
+            PQclear(result);
+            Py_INCREF(Py_None);
+            return Py_None;
+        default:
+            set_error_msg(InternalError, "Unknown result status");
+    }
+
+    PQclear(result);
+    return NULL; /* error detected on query */
+ }
+
+/* Base method for execution of all different kinds of queries */
+static PyObject *
+_conn_query(connObject *self, PyObject *args, int prepared, int async)
 {
     PyObject *query_str_obj, *param_obj = NULL;
     PGresult* result;
@@ -229,8 +277,8 @@ _conn_query(connObject *self, PyObject *args, int prepared)
         const char **parms, **p;
         register int i;
 
-        str = (PyObject **) PyMem_Malloc(nparms * sizeof(*str));
-        parms = (const char **) PyMem_Malloc(nparms * sizeof(*parms));
+        str = (PyObject **) PyMem_Malloc((size_t) nparms * sizeof(*str));
+        parms = (const char **) PyMem_Malloc((size_t) nparms * sizeof(*parms));
         if (!str || !parms) {
             PyMem_Free((void *) parms); PyMem_Free(str);
             Py_XDECREF(query_str_obj); Py_XDECREF(param_obj);
@@ -282,11 +330,19 @@ _conn_query(connObject *self, PyObject *args, int prepared)
         }
 
         Py_BEGIN_ALLOW_THREADS
-        result = prepared ?
-            PQexecPrepared(self->cnx, query, nparms,
-                parms, NULL, NULL, 0) :
-            PQexecParams(self->cnx, query, nparms,
-                NULL, parms, NULL, NULL, 0);
+        if (async) {
+            status = PQsendQueryParams(self->cnx, query, nparms,
+                NULL, (const char * const *)parms, NULL, NULL, 0);
+            result = NULL;
+        }
+        else {
+            result = prepared ?
+                PQexecPrepared(self->cnx, query, nparms,
+                    parms, NULL, NULL, 0) :
+                PQexecParams(self->cnx, query, nparms,
+                    NULL, parms, NULL, NULL, 0);
+            status = result != NULL;
+        }
         Py_END_ALLOW_THREADS
 
         PyMem_Free((void *) parms);
@@ -295,10 +351,17 @@ _conn_query(connObject *self, PyObject *args, int prepared)
     }
     else {
         Py_BEGIN_ALLOW_THREADS
-        result = prepared ?
-            PQexecPrepared(self->cnx, query, 0,
-                NULL, NULL, NULL, 0) :
-            PQexec(self->cnx, query);
+        if (async) {
+            status = PQsendQuery(self->cnx, query);
+            result = NULL;
+        }
+        else {
+            result = prepared ?
+                PQexecPrepared(self->cnx, query, 0,
+                    NULL, NULL, NULL, 0) :
+                PQexec(self->cnx, query);
+            status = result != NULL;
+        }
         Py_END_ALLOW_THREADS
     }
 
@@ -307,7 +370,7 @@ _conn_query(connObject *self, PyObject *args, int prepared)
     Py_XDECREF(param_obj);
 
     /* checks result validity */
-    if (!result) {
+    if (!status) {
         PyErr_SetString(PyExc_ValueError, PQerrorMessage(self->cnx));
         return NULL;
     }
@@ -317,66 +380,33 @@ _conn_query(connObject *self, PyObject *args, int prepared)
     self->date_format = date_format; /* this is normally NULL */
 
     /* checks result status */
-    if ((status = PQresultStatus(result)) != PGRES_TUPLES_OK) {
-        switch (status) {
-            case PGRES_EMPTY_QUERY:
-                PyErr_SetString(PyExc_ValueError, "Empty query");
-                break;
-            case PGRES_BAD_RESPONSE:
-            case PGRES_FATAL_ERROR:
-            case PGRES_NONFATAL_ERROR:
-                set_error(ProgrammingError, "Cannot execute query",
-                    self->cnx, result);
-                break;
-            case PGRES_COMMAND_OK:
-                {   /* INSERT, UPDATE, DELETE */
-                    Oid oid = PQoidValue(result);
+    if (result && (status = PQresultStatus(result)) != PGRES_TUPLES_OK)
+        return _conn_non_query_result(status, result, self->cnx);
 
-                    if (oid == InvalidOid) {  /* not a single insert */
-                        char *ret = PQcmdTuples(result);
-
-                        if (ret[0]) {  /* return number of rows affected */
-                            PyObject *obj = PyStr_FromString(ret);
-                            PQclear(result);
-                            return obj;
-                        }
-                        PQclear(result);
-                        Py_INCREF(Py_None);
-                        return Py_None;
-                    }
-                    /* for a single insert, return the oid */
-                    PQclear(result);
-                    return PyInt_FromLong(oid);
-                }
-            case PGRES_COPY_OUT: /* no data will be received */
-            case PGRES_COPY_IN:
-                PQclear(result);
-                Py_INCREF(Py_None);
-                return Py_None;
-            default:
-                set_error_msg(InternalError, "Unknown result status");
-        }
-
-        PQclear(result);
-        return NULL; /* error detected on query */
-    }
-
-    if (!(query_obj = PyObject_NEW(queryObject, &queryType)))
+    if (!(query_obj = PyObject_New(queryObject, &queryType)))
         return PyErr_NoMemory();
 
     /* stores result and returns object */
     Py_XINCREF(self);
     query_obj->pgcnx = self;
     query_obj->result = result;
+    query_obj->async = async;
     query_obj->encoding = encoding;
     query_obj->current_row = 0;
-    query_obj->max_row = PQntuples(result);
-    query_obj->num_fields = PQnfields(result);
-    query_obj->col_types = get_col_types(result, query_obj->num_fields);
-    if (!query_obj->col_types) {
-        Py_DECREF(query_obj);
-        Py_DECREF(self);
-        return NULL;
+    if (async) {
+        query_obj->max_row = 0;
+        query_obj->num_fields = 0;
+        query_obj->col_types = NULL;
+    }
+    else {
+        query_obj->max_row = PQntuples(result);
+        query_obj->num_fields = PQnfields(result);
+        query_obj->col_types = get_col_types(result, query_obj->num_fields);
+        if (!query_obj->col_types) {
+            Py_DECREF(query_obj);
+            Py_DECREF(self);
+            return NULL;
+        }
     }
 
     return (PyObject *) query_obj;
@@ -391,7 +421,19 @@ static char conn_query__doc__[] =
 static PyObject *
 conn_query(connObject *self, PyObject *args)
 {
-    return _conn_query(self, args, 0);
+    return _conn_query(self, args, 0, 0);
+}
+
+/* Asynchronous database query */
+static char conn_send_query__doc__[] =
+"send_query(sql, [arg]) -- create a new asynchronous query for this connection\n\n"
+"You must pass the SQL (string) request and you can optionally pass\n"
+"a tuple with positional parameters.\n";
+
+static PyObject *
+conn_send_query(connObject *self, PyObject *args)
+{
+    return _conn_query(self, args, 0, 1);
 }
 
 /* Execute prepared statement. */
@@ -403,7 +445,7 @@ static char conn_query_prepared__doc__[] =
 static PyObject *
 conn_query_prepared(connObject *self, PyObject *args)
 {
-    return _conn_query(self, args, 1);
+    return _conn_query(self, args, 1, 0);
 }
 
 /* Create prepared statement. */
@@ -478,7 +520,7 @@ conn_describe_prepared(connObject *self, PyObject *args)
     result = PQdescribePrepared(self->cnx, name);
     Py_END_ALLOW_THREADS
     if (result && PQresultStatus(result) == PGRES_COMMAND_OK) {
-        queryObject *query_obj = PyObject_NEW(queryObject, &queryType);
+        queryObject *query_obj = PyObject_New(queryObject, &queryType);
         if (!query_obj)
             return PyErr_NoMemory();
         Py_XINCREF(self);
@@ -583,13 +625,70 @@ conn_endcopy(connObject *self, PyObject *noargs)
     Py_INCREF(Py_None);
     return Py_None;
 }
+
+/* Direct access function: set blocking status. */
+static char conn_set_non_blocking__doc__[] =
+"set_non_blocking() -- set the non-blocking status of the connection";
+
+static PyObject *
+conn_set_non_blocking(connObject *self, PyObject *args)
+{
+    int non_blocking;
+
+    if (!self->cnx) {
+        PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+        return NULL;
+    }
+
+    if (!PyArg_ParseTuple(args, "i", &non_blocking)) {
+        PyErr_SetString(PyExc_TypeError, "setnonblocking(tf), with boolean.");
+        return NULL;
+    }
+
+    if (PQsetnonblocking(self->cnx, non_blocking) < 0) {
+        PyErr_SetString(PyExc_IOError, PQerrorMessage(self->cnx));
+        return NULL;
+    }
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+/* Direct access function: get blocking status. */
+static char conn_is_non_blocking__doc__[] =
+"is_non_blocking() -- report the blocking status of the connection";
+
+static PyObject *
+conn_is_non_blocking(connObject *self, PyObject *args)
+{
+    int rc;
+
+    if (!self->cnx) {
+        PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+        return NULL;
+    }
+
+    if (!PyArg_ParseTuple(args, "")) {
+        PyErr_SetString(PyExc_TypeError,
+            "method is_non_blocking() takes no parameters");
+        return NULL;
+    }
+
+    rc = PQisnonblocking(self->cnx);
+    if (rc < 0) {
+        PyErr_SetString(PyExc_IOError, PQerrorMessage(self->cnx));
+        return NULL;
+    }
+
+    return PyBool_FromLong(rc);
+}
 #endif /* DIRECT_ACCESS */
 
 
 /* Insert table */
 static char conn_inserttable__doc__[] =
-"inserttable(table, data) -- insert list into table\n\n"
-"The fields in the list must be in the same order as in the table.\n";
+"inserttable(table, data, [columns]) -- insert list into table\n\n"
+"The fields in the list must be in the same order as in the table\n"
+"or in the list of columns if one is specified.\n";
 
 static PyObject *
 conn_inserttable(connObject *self, PyObject *args)
@@ -598,10 +697,11 @@ conn_inserttable(connObject *self, PyObject *args)
     char *table, *buffer, *bufpt;
     int encoding;
     size_t bufsiz;
-    PyObject *list, *sublist, *item;
+    PyObject *list, *sublist, *item, *columns = NULL;
     PyObject *(*getitem) (PyObject *, Py_ssize_t);
     PyObject *(*getsubitem) (PyObject *, Py_ssize_t);
-    Py_ssize_t i, j, m, n;
+    PyObject *(*getcolumn) (PyObject *, Py_ssize_t);
+    Py_ssize_t i, j, m, n = 0;
 
     if (!self->cnx) {
         PyErr_SetString(PyExc_TypeError, "Connection is not valid");
@@ -609,7 +709,7 @@ conn_inserttable(connObject *self, PyObject *args)
     }
 
     /* gets arguments */
-    if (!PyArg_ParseTuple(args, "sO:filter", &table, &list)) {
+    if (!PyArg_ParseTuple(args, "sO|O", &table, &list, &columns)) {
         PyErr_SetString(
             PyExc_TypeError,
             "Method inserttable() expects a string and a list as arguments");
@@ -633,12 +733,68 @@ conn_inserttable(connObject *self, PyObject *args)
         return NULL;
     }
 
+    /* checks columns type */
+    if (columns) {
+        if (PyList_Check(columns)) {
+            n = PyList_Size(columns);
+            getcolumn = PyList_GetItem;
+        }
+        else if (PyTuple_Check(columns)) {
+            n = PyTuple_Size(columns);
+            getcolumn = PyTuple_GetItem;
+        }
+        else {
+            PyErr_SetString(
+                PyExc_TypeError,
+                "Method inserttable() expects a list or a tuple"
+                " as third argument");
+            return NULL;
+        }
+        if (!n) {
+            /* no columns specified, nothing to do */
+            Py_INCREF(Py_None);
+            return Py_None;
+        }
+    }
+
     /* allocate buffer */
     if (!(buffer = PyMem_Malloc(MAX_BUFFER_SIZE)))
         return PyErr_NoMemory();
 
+    encoding = PQclientEncoding(self->cnx);
+
     /* starts query */
-    sprintf(buffer, "copy %s from stdin", table);
+    bufpt = buffer;
+    table = PQescapeIdentifier(self->cnx, table, strlen(table));
+    bufpt += sprintf(bufpt, "copy %s", table);
+    PQfreemem(table);
+    if (columns) {
+        /* adds a string like f" ({','.join(columns)})" */
+        bufpt += sprintf(bufpt, " (");
+        for (j = 0; j < n; ++j) {
+            PyObject *obj = getcolumn(columns, j);
+            ssize_t slen;
+            char *col;
+
+            if (PyBytes_Check(obj)) {
+                PyBytes_AsStringAndSize(obj, &col, &slen);
+            }
+            else if (PyUnicode_Check(obj)) {
+                obj = get_encoded_string(obj, encoding);
+                if (!obj) return NULL; /* pass the UnicodeEncodeError */
+                PyBytes_AsStringAndSize(obj, &col, &slen);
+                Py_DECREF(obj);
+            } else {
+                PyErr_SetString(
+                    PyExc_TypeError,
+                    "The third argument must contain only strings");
+            }
+            col = PQescapeIdentifier(self->cnx, col, (size_t) slen);
+            bufpt += sprintf(bufpt, "%s%s", col, j == n - 1 ? ")" : ",");
+            PQfreemem(col);
+        }
+    }
+    sprintf(bufpt, " from stdin");
 
     Py_BEGIN_ALLOW_THREADS
     result = PQexec(self->cnx, buffer);
@@ -650,12 +806,8 @@ conn_inserttable(connObject *self, PyObject *args)
         return NULL;
     }
 
-    encoding = PQclientEncoding(self->cnx);
-
     PQclear(result);
 
-    n = 0; /* not strictly necessary but avoids warning */
-
     /* feed table */
     for (i = 0; i < m; ++i) {
         sublist = getitem(list, i);
@@ -906,9 +1058,9 @@ conn_escape_literal(connObject *self, PyObject *string)
     Py_XDECREF(tmp_obj);
 
     if (encoding == -1)
-        to_obj = PyBytes_FromStringAndSize(to, to_length);
+        to_obj = PyBytes_FromStringAndSize(to, (Py_ssize_t) to_length);
     else
-        to_obj = get_decoded_string(to, to_length, encoding);
+        to_obj = get_decoded_string(to, (Py_ssize_t) to_length, encoding);
     if (to)
         PQfreemem(to);
     return to_obj;
@@ -951,9 +1103,9 @@ conn_escape_identifier(connObject *self, PyObject *string)
     Py_XDECREF(tmp_obj);
 
     if (encoding == -1)
-        to_obj = PyBytes_FromStringAndSize(to, to_length);
+        to_obj = PyBytes_FromStringAndSize(to, (Py_ssize_t) to_length);
     else
-        to_obj = get_decoded_string(to, to_length, encoding);
+        to_obj = get_decoded_string(to, (Py_ssize_t) to_length, encoding);
     if (to)
         PQfreemem(to);
     return to_obj;
@@ -992,9 +1144,9 @@ conn_escape_string(connObject *self, PyObject *string)
         return NULL;
     }
 
-    to_length = 2*from_length + 1;
+    to_length = 2 * (size_t) from_length + 1;
     if ((Py_ssize_t) to_length < from_length) { /* overflow */
-        to_length = from_length;
+        to_length = (size_t) from_length;
         from_length = (from_length - 1)/2;
     }
     to = (char *) PyMem_Malloc(to_length);
@@ -1004,9 +1156,9 @@ conn_escape_string(connObject *self, PyObject *string)
     Py_XDECREF(tmp_obj);
 
     if (encoding == -1)
-        to_obj = PyBytes_FromStringAndSize(to, to_length);
+        to_obj = PyBytes_FromStringAndSize(to, (Py_ssize_t) to_length);
     else
-        to_obj = get_decoded_string(to, to_length, encoding);
+        to_obj = get_decoded_string(to, (Py_ssize_t) to_length, encoding);
     PyMem_Free(to);
     return to_obj;
 }
@@ -1048,9 +1200,9 @@ conn_escape_bytea(connObject *self, PyObject *data)
     Py_XDECREF(tmp_obj);
 
     if (encoding == -1)
-        to_obj = PyBytes_FromStringAndSize(to, to_length - 1);
+        to_obj = PyBytes_FromStringAndSize(to, (Py_ssize_t) to_length - 1);
     else
-        to_obj = get_decoded_string(to, to_length - 1, encoding);
+        to_obj = get_decoded_string(to, (Py_ssize_t) to_length - 1, encoding);
     if (to)
         PQfreemem(to);
     return to_obj;
@@ -1064,7 +1216,7 @@ large_new(connObject *pgcnx, Oid oid)
 {
     largeObject *large_obj;
 
-    if (!(large_obj = PyObject_NEW(largeObject, &largeType))) {
+    if (!(large_obj = PyObject_New(largeObject, &largeType))) {
         return NULL;
     }
 
@@ -1269,6 +1421,40 @@ conn_get_cast_hook(connObject *self, PyObject *noargs)
     return ret;
 }
 
+/* Get asynchronous connection state. */
+static char conn_poll__doc__[] =
+"poll() -- Completes an asynchronous connection";
+
+static PyObject *
+conn_poll(connObject *self, PyObject *args)
+{
+    int rc;
+
+    if (!self->cnx) {
+        PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+        return NULL;
+    }
+
+    /* check args */
+    if (!PyArg_ParseTuple(args, "")) {
+        PyErr_SetString(PyExc_TypeError,
+            "method poll() takes no parameters");
+        return NULL;
+    }
+
+    Py_BEGIN_ALLOW_THREADS
+    rc = PQconnectPoll(self->cnx);
+    Py_END_ALLOW_THREADS
+
+    if (rc == PGRES_POLLING_FAILED) {
+        set_error(InternalError, "Polling failed", self->cnx, NULL);
+        Py_XDECREF(self);
+        return NULL;
+    }
+
+    return PyInt_FromLong(rc);
+}
+
 /* Set notice receiver callback function. */
 static char conn_set_notice_receiver__doc__[] =
 "set_notice_receiver(func) -- set the current notice receiver";
@@ -1417,12 +1603,16 @@ static struct PyMethodDef conn_methods[] = {
         METH_NOARGS, conn_source__doc__},
     {"query", (PyCFunction) conn_query,
         METH_VARARGS, conn_query__doc__},
+    {"send_query", (PyCFunction) conn_send_query,
+        METH_VARARGS, conn_send_query__doc__},
     {"query_prepared", (PyCFunction) conn_query_prepared,
         METH_VARARGS, conn_query_prepared__doc__},
     {"prepare", (PyCFunction) conn_prepare,
         METH_VARARGS, conn_prepare__doc__},
     {"describe_prepared", (PyCFunction) conn_describe_prepared,
         METH_VARARGS, conn_describe_prepared__doc__},
+    {"poll", (PyCFunction) conn_poll,
+        METH_VARARGS, conn_poll__doc__},
     {"reset", (PyCFunction) conn_reset,
         METH_NOARGS, conn_reset__doc__},
     {"cancel", (PyCFunction) conn_cancel,
@@ -1468,6 +1658,10 @@ static struct PyMethodDef conn_methods[] = {
         METH_NOARGS, conn_getline__doc__},
     {"endcopy", (PyCFunction) conn_endcopy,
         METH_NOARGS, conn_endcopy__doc__},
+    {"set_non_blocking", (PyCFunction) conn_set_non_blocking,
+        METH_O, conn_set_non_blocking__doc__},
+    {"is_non_blocking", (PyCFunction) conn_is_non_blocking,
+        METH_NOARGS, conn_is_non_blocking__doc__},
 #endif /* DIRECT_ACCESS */
 
 #ifdef LARGE_OBJECTS
