@@ -12,8 +12,12 @@
 static void
 large_dealloc(largeObject *self)
 {
-    if (self->lo_fd >= 0 && self->pgcnx->valid)
-        lo_close(self->pgcnx->cnx, self->lo_fd);
+    /* Note: We do not try to close the large object here anymore,
+       since the server automatically closes it at the end of the
+       transaction in which it was created. So the object might already
+       be closed, which will then cause error messages on the server.
+       In other situations we might close the object too early here
+       if the Python object falls out of scope but is still needed. */
 
     Py_XDECREF(self->pgcnx);
     PyObject_Del(self);
@@ -81,7 +85,7 @@ large_getattr(largeObject *self, PyObject *nameobj)
     /* large object oid */
     if (!strcmp(name, "oid")) {
         if (_check_lo_obj(self, 0))
-            return PyInt_FromLong(self->lo_oid);
+            return PyInt_FromLong((long) self->lo_oid);
         PyErr_Clear();
         Py_INCREF(Py_None);
         return Py_None;
@@ -199,7 +203,7 @@ large_read(largeObject *self, PyObject *args)
     buffer = PyBytes_FromStringAndSize((char *) NULL, size);
 
     if ((size = lo_read(self->pgcnx->cnx, self->lo_fd,
-        PyBytes_AS_STRING((PyBytesObject *) (buffer)), size)) == -1)
+        PyBytes_AS_STRING((PyBytesObject *) (buffer)), (size_t) size)) == -1)
     {
         PyErr_SetString(PyExc_IOError, "Error while reading");
         Py_XDECREF(buffer);
@@ -237,7 +241,7 @@ large_write(largeObject *self, PyObject *args)
 
     /* sends query */
     if ((size = lo_write(self->pgcnx->cnx, self->lo_fd, buffer,
-                         (int)bufsize)) != bufsize)
+                         (size_t) bufsize)) != bufsize)
     {
         PyErr_SetString(PyExc_IOError, "Buffer truncated during write");
         return NULL;
