@@ -102,6 +102,89 @@ _query_row_as_tuple(queryObject *self)
     return row_tuple;
 }
 
+/* Fetch the result if this is an asynchronous query and it has not yet
+   been fetched in this round-trip. Also mark whether the result should
+   be kept for this round-trip (e.g. to be used in an iterator).
+   If this is a normal query result, the query itself will be returned,
+   otherwise a result value will be returned that shall be passed on. */
+static PyObject *
+_get_async_result(queryObject *self, int keep) {
+    int fetch = 0;
+
+    if (self->async) {
+        if (self->async == 1) {
+            fetch = 1;
+            if (keep) {
+                /* mark query as fetched, do not fetch again */
+                self->async = 2;
+            }
+        } else if (!keep) {
+            self->async = 1;
+        }
+    }
+
+    if (fetch) {
+        int status;
+
+        if (!self->pgcnx) {
+            PyErr_SetString(PyExc_TypeError, "Connection is not valid");
+            return NULL;
+        }
+
+        Py_BEGIN_ALLOW_THREADS
+        if (self->result) {
+            PQclear(self->result);
+        }
+        self->result = PQgetResult(self->pgcnx->cnx);
+        Py_END_ALLOW_THREADS
+        if (!self->result) {
+            /* end of result set, return None */
+            Py_DECREF(self->pgcnx);
+            self->pgcnx = NULL;
+            Py_INCREF(Py_None);
+            return Py_None;
+        }
+
+        if ((status = PQresultStatus(self->result)) != PGRES_TUPLES_OK) {
+            PyObject* result = _conn_non_query_result(
+                status, self->result, self->pgcnx->cnx);
+            self->result = NULL; /* since this has been already cleared */
+            if (!result) {
+                /* Raise an error. We need to call PQgetResult() to clear the
+                   connection state. This should return NULL the first time. */
+                self->result = PQgetResult(self->pgcnx->cnx);
+                while (self->result) {
+                    PQclear(self->result);
+                    self->result = PQgetResult(self->pgcnx->cnx);
+                    Py_DECREF(self->pgcnx);
+                    self->pgcnx = NULL;
+                }
+            }
+            else if (result == Py_None) {
+                /* It's would be confusing to return None here because the
+                   caller has to call again until we return None. We can't
+                   just consume that final None because we don't know if there
+                   are additional statements following this one, so we return
+                   an empty string where query() would return None. */
+                Py_DECREF(result);
+                result = PyStr_FromString("");
+            }
+            return result;
+        }
+
+        self->max_row = PQntuples(self->result);
+        self->num_fields = PQnfields(self->result);
+        self->col_types = get_col_types(self->result, self->num_fields);
+        if (!self->col_types) {
+            Py_DECREF(self);
+            Py_DECREF(self);
+            return NULL;
+        }
+    }
+    /* return the query object itself as sentinel for a normal query result */
+    return (PyObject *)self;
+}
+
 /* Return given item from a query object. */
 static PyObject *
 query_getitem(PyObject *self, Py_ssize_t i)
@@ -110,7 +193,10 @@ query_getitem(PyObject *self, Py_ssize_t i)
     PyObject *tmp;
     long row;
 
-    tmp = PyLong_FromSize_t(i);
+    if ((tmp = _get_async_result(q, 0)) != (PyObject *)self)
+        return tmp;
+
+    tmp = PyLong_FromSize_t((size_t) i);
     row = PyLong_AsLong(tmp);
     Py_DECREF(tmp);
 
@@ -119,7 +205,7 @@ query_getitem(PyObject *self, Py_ssize_t i)
         return NULL;
     }
 
-    q->current_row = row;
+    q->current_row = (int) row;
     return _query_row_as_tuple(q);
 }
 
@@ -127,6 +213,11 @@ query_getitem(PyObject *self, Py_ssize_t i)
    Returns the default iterator yielding rows as tuples. */
 static PyObject* query_iter(queryObject *self)
 {
+    PyObject *res;
+
+    if ((res = _get_async_result(self, 0)) != (PyObject *)self)
+        return res;
+
     self->current_row = 0;
     Py_INCREF(self);
     return (PyObject*) self;
@@ -149,6 +240,19 @@ query_next(queryObject *self, PyObject *noargs)
     return row_tuple;
 }
 
+#ifdef MEMORY_SIZE
+
+/* Get number of bytes allocated for PGresult object */
+static char query_memsize__doc__[] =
+"memsize() -- return number of bytes allocated by query result";
+static PyObject *
+query_memsize(queryObject *self, PyObject *noargs)
+{
+    return PyLong_FromSize_t(PQresultMemorySize(self->result));
+}
+
+#endif /* MEMORY_SIZE */
+
 /* Get number of rows. */
 static char query_ntuples__doc__[] =
 "ntuples() -- return number of tuples returned by query";
@@ -236,6 +340,77 @@ query_fieldnum(queryObject *self, PyObject *args)
     return PyInt_FromLong(num);
 }
 
+/* Build a tuple with info for query field with given number. */
+static PyObject *
+_query_build_field_info(PGresult *res, int col_num) {
+    PyObject *info;
+
+    info = PyTuple_New(4);
+    if (info) {
+        PyTuple_SET_ITEM(info, 0, PyStr_FromString(PQfname(res, col_num)));
+        PyTuple_SET_ITEM(info, 1, PyInt_FromLong((long) PQftype(res, col_num)));
+        PyTuple_SET_ITEM(info, 2, PyInt_FromLong(PQfsize(res, col_num)));
+        PyTuple_SET_ITEM(info, 3, PyInt_FromLong(PQfmod(res, col_num)));
+    }
+    return info;
+}
+
+/* Get information on one or all fields of the query result. */
+static char query_fieldinfo__doc__[] =
+"fieldinfo() -- return info on field(s) in query";
+
+static PyObject *
+query_fieldinfo(queryObject *self, PyObject *args)
+{
+    PyObject *result, *field = NULL;
+    int num;
+
+    /* gets args */
+    if (!PyArg_ParseTuple(args, "|O", &field)) {
+        PyErr_SetString(PyExc_TypeError,
+                        "Method fieldinfo() takes one optional argument only");
+        return NULL;
+    }
+
+    /* check optional field arg */
+    if (field) {
+        /* gets field number */
+        if (PyBytes_Check(field)) {
+            num = PQfnumber(self->result, PyBytes_AsString(field));
+        } else if (PyStr_Check(field)) {
+            PyObject *tmp = get_encoded_string(field, self->encoding);
+            if (!tmp) return NULL;
+            num = PQfnumber(self->result, PyBytes_AsString(tmp));
+            Py_DECREF(tmp);
+        } else if (PyInt_Check(field)) {
+            num = (int) PyInt_AsLong(field);
+        } else {
+            PyErr_SetString(PyExc_TypeError,
+                            "Field should be given as column number or name");
+            return NULL;
+        }
+        if (num < 0 || num >= self->num_fields) {
+            PyErr_SetString(PyExc_IndexError, "Unknown field");
+            return NULL;
+        }
+        return _query_build_field_info(self->result, num);
+    }
+
+    if (!(result = PyTuple_New(self->num_fields))) {
+        return NULL;
+    }
+    for (num = 0; num < self->num_fields; ++num) {
+        PyObject *info = _query_build_field_info(self->result, num);
+        if (!info) {
+            Py_DECREF(result);
+            return NULL;
+        }
+        PyTuple_SET_ITEM(result, num, info);
+    }
+    return result;
+}
+
+
 /* Retrieve one row from the result as a tuple. */
 static char query_one__doc__[] =
 "one() -- Get one row from the result of a query\n\n"
@@ -248,12 +423,16 @@ query_one(queryObject *self, PyObject *noargs)
 {
     PyObject *row_tuple;
 
-    if (self->current_row >= self->max_row) {
-        Py_INCREF(Py_None); return Py_None;
+    if ((row_tuple = _get_async_result(self, 0)) == (PyObject *)self) {
+
+        if (self->current_row >= self->max_row) {
+            Py_INCREF(Py_None); return Py_None;
+        }
+
+        row_tuple = _query_row_as_tuple(self);
+        if (row_tuple) ++self->current_row;
     }
 
-    row_tuple = _query_row_as_tuple(self);
-    if (row_tuple) ++self->current_row;
     return row_tuple;
 }
 
@@ -270,17 +449,21 @@ query_single(queryObject *self, PyObject *noargs)
 {
     PyObject *row_tuple;
 
-    if (self->max_row != 1) {
-        if (self->max_row)
-            set_error_msg(MultipleResultsError, "Multiple results found");
-        else
-            set_error_msg(NoResultError, "No result found");
-        return NULL;
+    if ((row_tuple = _get_async_result(self, 0)) == (PyObject *)self) {
+
+        if (self->max_row != 1) {
+            if (self->max_row)
+                set_error_msg(MultipleResultsError, "Multiple results found");
+            else
+                set_error_msg(NoResultError, "No result found");
+            return NULL;
+        }
+
+        self->current_row = 0;
+        row_tuple = _query_row_as_tuple(self);
+        if (row_tuple) ++self->current_row;
     }
 
-    self->current_row = 0;
-    row_tuple = _query_row_as_tuple(self);
-    if (row_tuple) ++self->current_row;
     return row_tuple;
 }
 
@@ -296,17 +479,20 @@ query_getresult(queryObject *self, PyObject *noargs)
     PyObject *result_list;
     int i;
 
-    if (!(result_list = PyList_New(self->max_row))) {
-        return NULL;
-    }
+    if ((result_list = _get_async_result(self, 0)) == (PyObject *)self) {
 
-    for (i = self->current_row = 0; i < self->max_row; ++i) {
-        PyObject *row_tuple = query_next(self, noargs);
+        if (!(result_list = PyList_New(self->max_row))) {
+            return NULL;
+        }
 
-        if (!row_tuple) {
-            Py_DECREF(result_list); return NULL;
+        for (i = self->current_row = 0; i < self->max_row; ++i) {
+            PyObject *row_tuple = query_next(self, noargs);
+
+            if (!row_tuple) {
+                Py_DECREF(result_list); return NULL;
+            }
+            PyList_SET_ITEM(result_list, i, row_tuple);
         }
-        PyList_SET_ITEM(result_list, i, row_tuple);
     }
 
     return result_list;
@@ -365,12 +551,16 @@ query_onedict(queryObject *self, PyObject *noargs)
 {
     PyObject *row_dict;
 
-    if (self->current_row >= self->max_row) {
-        Py_INCREF(Py_None); return Py_None;
+    if ((row_dict = _get_async_result(self, 0)) == (PyObject *)self) {
+
+        if (self->current_row >= self->max_row) {
+            Py_INCREF(Py_None); return Py_None;
+        }
+
+        row_dict = _query_row_as_dict(self);
+        if (row_dict) ++self->current_row;
     }
 
-    row_dict = _query_row_as_dict(self);
-    if (row_dict) ++self->current_row;
     return row_dict;
 }
 
@@ -388,17 +578,21 @@ query_singledict(queryObject *self, PyObject *noargs)
 {
     PyObject *row_dict;
 
-    if (self->max_row != 1) {
-        if (self->max_row)
-            set_error_msg(MultipleResultsError, "Multiple results found");
-        else
-            set_error_msg(NoResultError, "No result found");
-        return NULL;
+    if ((row_dict = _get_async_result(self, 0)) == (PyObject *)self) {
+
+        if (self->max_row != 1) {
+            if (self->max_row)
+                set_error_msg(MultipleResultsError, "Multiple results found");
+            else
+                set_error_msg(NoResultError, "No result found");
+            return NULL;
+        }
+
+        self->current_row = 0;
+        row_dict = _query_row_as_dict(self);
+        if (row_dict) ++self->current_row;
     }
 
-    self->current_row = 0;
-    row_dict = _query_row_as_dict(self);
-    if (row_dict) ++self->current_row;
     return row_dict;
 }
 
@@ -414,17 +608,20 @@ query_dictresult(queryObject *self, PyObject *noargs)
     PyObject *result_list;
     int i;
 
-    if (!(result_list = PyList_New(self->max_row))) {
-        return NULL;
-    }
+    if ((result_list = _get_async_result(self, 0)) == (PyObject *)self) {
 
-    for (i = self->current_row = 0; i < self->max_row; ++i) {
-        PyObject *row_dict = query_next_dict(self, noargs);
+        if (!(result_list = PyList_New(self->max_row))) {
+            return NULL;
+        }
 
-        if (!row_dict) {
-            Py_DECREF(result_list); return NULL;
+        for (i = self->current_row = 0; i < self->max_row; ++i) {
+            PyObject *row_dict = query_next_dict(self, noargs);
+
+            if (!row_dict) {
+                Py_DECREF(result_list); return NULL;
+            }
+            PyList_SET_ITEM(result_list, i, row_dict);
         }
-        PyList_SET_ITEM(result_list, i, row_dict);
     }
 
     return result_list;
@@ -439,10 +636,15 @@ static char query_dictiter__doc__[] =
 static PyObject *
 query_dictiter(queryObject *self, PyObject *noargs)
 {
+    PyObject *res;
+
     if (!dictiter) {
         return query_dictresult(self, noargs);
     }
 
+    if ((res = _get_async_result(self, 1)) != (PyObject *)self)
+        return res;
+
     return PyObject_CallFunction(dictiter, "(O)", self);
 }
 
@@ -456,10 +658,15 @@ static char query_onenamed__doc__[] =
 static PyObject *
 query_onenamed(queryObject *self, PyObject *noargs)
 {
+    PyObject *res;
+
     if (!namednext) {
         return query_one(self, noargs);
     }
 
+    if ((res = _get_async_result(self, 1)) != (PyObject *)self)
+        return res;
+
     if (self->current_row >= self->max_row) {
         Py_INCREF(Py_None); return Py_None;
     }
@@ -478,10 +685,15 @@ static char query_singlenamed__doc__[] =
 static PyObject *
 query_singlenamed(queryObject *self, PyObject *noargs)
 {
+    PyObject *res;
+
     if (!namednext) {
         return query_single(self, noargs);
     }
 
+    if ((res = _get_async_result(self, 1)) != (PyObject *)self)
+        return res;
+
     if (self->max_row != 1) {
         if (self->max_row)
             set_error_msg(MultipleResultsError, "Multiple results found");
@@ -509,11 +721,15 @@ query_namedresult(queryObject *self, PyObject *noargs)
         return query_getresult(self, noargs);
     }
 
-    res = PyObject_CallFunction(namediter, "(O)", self);
-    if (!res) return NULL;
-    if (PyList_Check(res)) return res;
-    res_list = PySequence_List(res);
-    Py_DECREF(res);
+    if ((res_list = _get_async_result(self, 1)) == (PyObject *)self) {
+
+        res = PyObject_CallFunction(namediter, "(O)", self);
+        if (!res) return NULL;
+        if (PyList_Check(res)) return res;
+        res_list = PySequence_List(res);
+        Py_DECREF(res);
+    }
+
     return res_list;
 }
 
@@ -532,11 +748,15 @@ query_namediter(queryObject *self, PyObject *noargs)
         return query_iter(self);
     }
 
-    res = PyObject_CallFunction(namediter, "(O)", self);
-    if (!res) return NULL;
-    if (!PyList_Check(res)) return res;
-    res_iter = (Py_TYPE(res)->tp_iter)((PyObject *) self);
-    Py_DECREF(res);
+    if ((res_iter = _get_async_result(self, 1)) == (PyObject *)self) {
+
+        res = PyObject_CallFunction(namediter, "(O)", self);
+        if (!res) return NULL;
+        if (!PyList_Check(res)) return res;
+        res_iter = (Py_TYPE(res)->tp_iter)((PyObject *) self);
+        Py_DECREF(res);
+    }
+
     return res_iter;
 }
 
@@ -551,25 +771,28 @@ query_scalarresult(queryObject *self, PyObject *noargs)
 {
     PyObject *result_list;
 
-    if (!self->num_fields) {
-        set_error_msg(ProgrammingError, "No fields in result");
-        return NULL;
-    }
+    if ((result_list = _get_async_result(self, 0)) == (PyObject *)self) {
 
-    if (!(result_list = PyList_New(self->max_row))) {
-        return NULL;
-    }
+        if (!self->num_fields) {
+            set_error_msg(ProgrammingError, "No fields in result");
+            return NULL;
+        }
 
-    for (self->current_row = 0;
-         self->current_row < self->max_row;
-         ++self->current_row)
-    {
-        PyObject *value = _query_value_in_column(self, 0);
+        if (!(result_list = PyList_New(self->max_row))) {
+            return NULL;
+        }
 
-        if (!value) {
-            Py_DECREF(result_list); return NULL;
+        for (self->current_row = 0;
+             self->current_row < self->max_row;
+             ++self->current_row)
+        {
+            PyObject *value = _query_value_in_column(self, 0);
+
+            if (!value) {
+                Py_DECREF(result_list); return NULL;
+            }
+            PyList_SET_ITEM(result_list, self->current_row, value);
         }
-        PyList_SET_ITEM(result_list, self->current_row, value);
     }
 
     return result_list;
@@ -584,10 +807,15 @@ static char query_scalariter__doc__[] =
 static PyObject *
 query_scalariter(queryObject *self, PyObject *noargs)
 {
+    PyObject *res;
+
     if (!scalariter) {
         return query_scalarresult(self, noargs);
     }
 
+    if ((res = _get_async_result(self, 1)) != (PyObject *)self)
+        return res;
+
     if (!self->num_fields) {
         set_error_msg(ProgrammingError, "No fields in result");
         return NULL;
@@ -608,17 +836,21 @@ query_onescalar(queryObject *self, PyObject *noargs)
 {
     PyObject *value;
 
-    if (!self->num_fields) {
-        set_error_msg(ProgrammingError, "No fields in result");
-        return NULL;
+    if ((value = _get_async_result(self, 0)) == (PyObject *)self) {
+
+        if (!self->num_fields) {
+            set_error_msg(ProgrammingError, "No fields in result");
+            return NULL;
+        }
+
+        if (self->current_row >= self->max_row) {
+            Py_INCREF(Py_None); return Py_None;
+        }
+
+        value = _query_value_in_column(self, 0);
+        if (value) ++self->current_row;
     }
 
-    if (self->current_row >= self->max_row) {
-        Py_INCREF(Py_None); return Py_None;
-    }
-
-    value = _query_value_in_column(self, 0);
-    if (value) ++self->current_row;
     return value;
 }
 
@@ -635,22 +867,26 @@ query_singlescalar(queryObject *self, PyObject *noargs)
 {
     PyObject *value;
 
-    if (!self->num_fields) {
-        set_error_msg(ProgrammingError, "No fields in result");
-        return NULL;
+    if ((value = _get_async_result(self, 0)) == (PyObject *)self) {
+
+        if (!self->num_fields) {
+            set_error_msg(ProgrammingError, "No fields in result");
+            return NULL;
+        }
+
+        if (self->max_row != 1) {
+            if (self->max_row)
+                set_error_msg(MultipleResultsError, "Multiple results found");
+            else
+                set_error_msg(NoResultError, "No result found");
+            return NULL;
+        }
+
+        self->current_row = 0;
+        value = _query_value_in_column(self, 0);
+        if (value) ++self->current_row;
     }
 
-    if (self->max_row != 1) {
-        if (self->max_row)
-            set_error_msg(MultipleResultsError, "Multiple results found");
-        else
-            set_error_msg(NoResultError, "No result found");
-        return NULL;
-    }
-
-    self->current_row = 0;
-    value = _query_value_in_column(self, 0);
-    if (value) ++self->current_row;
     return value;
 }
 
@@ -704,8 +940,14 @@ static struct PyMethodDef query_methods[] = {
         METH_VARARGS, query_fieldnum__doc__},
     {"listfields", (PyCFunction) query_listfields,
         METH_NOARGS, query_listfields__doc__},
+    {"fieldinfo", (PyCFunction) query_fieldinfo,
+        METH_VARARGS, query_fieldinfo__doc__},
     {"ntuples", (PyCFunction) query_ntuples,
         METH_NOARGS, query_ntuples__doc__},
+#ifdef MEMORY_SIZE
+    {"memsize", (PyCFunction) query_memsize,
+        METH_NOARGS, query_memsize__doc__},
+#endif /* MEMORY_SIZE */
     {NULL, NULL}
 };
 
@@ -735,7 +977,7 @@ static PyTypeObject queryType = {
     0,                           /* tp_as_buffer */
     Py_TPFLAGS_DEFAULT
         |Py_TPFLAGS_HAVE_ITER,   /* tp_flags */
-    query__doc__,               /* tp_doc */
+    query__doc__,                /* tp_doc */
     0,                           /* tp_traverse */
     0,                           /* tp_clear */
     0,                           /* tp_richcompare */
